==> numerous_api_client-0.8.2.tar.gz: numerous_api_client-0.8.2/src/numerous_api_client/client/data_source.py <==
import enum
import threading
from datetime import datetime
from queue import Empty, Queue
from time import time
from typing import Any

import numpy as np
from scipy.interpolate import interp1d

from numerous_api_client.client.common import Interp, log


class DataSourceEmptyDataset(Exception):
    pass


class DataSourceHibernating(Exception):
    pass


class DataSourceCompleted(Exception):
    pass


class DataSourceStreamClosed(Exception):
    pass


class SourceStatus(enum.Enum):
    EMPTYDATASET = 0
    HIBERNATING = 1
    COMPLETED = 2
    STREAMCLOSED = 3
    OK = 4
    NODATA = 5


class StaticSource:
    def __init__(self, row_data, t0=0, t_end=None, dt=3600, generate_index=True):
        self.row = row_data
        self.source_type = 'static'
        self.t = np.inf
        self.t0 = t0
        self.subscribe = False
        self.generate_index = generate_index

    def get_at_time(self, t):
        self.t = t
        self.row['_index'] = t
        self.row['_index_relative'] = t - self.t0
        return self.row


class DynamicSource:
    def __init__(self, client: Any, scenario, execution, spec, t0=0, te=0, timeout=None, source_type=None,
                 interpolation_method=Interp.zero):
        self.timeout = timeout
        self.t0 = t0
        if te is None:
            te = 0
        self.te = te
        self.meta = client.get_timeseries_meta_data(scenario=scenario, execution=execution)
        self.closed = client.get_dataset_closed(scenario=scenario, execution=execution)
        self.name = scenario
        self.spec = spec
        self.spec['offset'] = self.spec['offset'] if 'offset' in self.spec else 0

        # The offset set in spec is the full offset from 0 - not relative from meta.offset
        offset = self.spec['offset']
        self.offset = offset
        self.execution = execution
        self.client = client

        # start time in original absolute time/input base time

        start = t0 - offset
        end = max(te - offset, 0)

        stats = client.get_timeseries_stats(None, scenario, execution)
        # If data set is open, we will try to listen for data as its being generated if t end not reached.
        self.subscribe = not self.closed

        if stats.min + offset > t0:
            raise Exception(f'dynamic input scenario {scenario} with tags {[t.name for t in self.meta.tags]} '
                            f'is offset to the future {stats.min + offset}>{t0}')

        self.start = start
        self.end = end

        print('scenario: ', scenario)
        print('scenario data is closed: ', self.closed)
        print('Offset in spec: ', datetime.fromtimestamp(spec['offset']))
        print('Offset in meta: ', datetime.fromtimestamp(self.meta.offset))
        print('meta off: ', datetime.fromtimestamp(self.meta.offset))
        print('Data stats: ', client.get_timeseries_stats(project=None, scenario=scenario, execution=execution))
        print('Combined offset: ', offset)
        print('Combined offset: ', datetime.fromtimestamp(offset))
        print('t0: ', datetime.fromtimestamp(t0))
        print('te: ', datetime.fromtimestamp(te))
        print('start: ', datetime.fromtimestamp(start))
        print('end: ', datetime.fromtimestamp(end))
        print('start: ', start)
        print('end: ', end)

        tags = set([s['tag'] for s in spec['tags'].values()])
        self.tags = tags

        self.generator = client.read_data_as_row(scenario=scenario, tags=tags, execution=execution, start=start,
                                                 end=end, subscribe=self.subscribe)
        self.source_type = source_type
        self.row_queue: Queue = Queue(100)
        self.t = -1
        # Keep a counter of rows read from data source - if nothing is read we don't want to requery.
        self._rows_read = 0
        self.row_buffer = RowBuffer(tags=[s for s in spec['tags'].keys()], offset=offset, method=interpolation_method)
        # Keep the last/presumably max ix read on record - if in continuous mode this needs to be added to the requery
        # as an offset.
        self.max_ix_read = 0
        self.last_get = None
        self.status = SourceStatus.OK

        def _next():
            complete = False
            lastdt = 0
            lastrow_index = None
            status = SourceStatus.STREAMCLOSED

            while not complete:
                for row_ in self.generator:
                    if row_ == "HIBERNATE":
                        status = SourceStatus.HIBERNATING
                        break
                    elif row_ == "STOP":
                        status = SourceStatus.COMPLETED
                        break

                    self._rows_read += 1
                    self.row_queue.put(row_)
                    if lastrow_index is None:
                        lastrow_index = row_['_index']
                    lastdt = row_['_index']-lastrow_index
                    lastrow_index = row_['_index']
                    self.max_ix_read = row_['_index']

                log.debug("%s: All data read: %s of %d to %d, read n rows: %d", scenario, self.t-offset, start, end,
                          self._rows_read)

                if self._rows_read <= 0:
                    # If nothing was returned the data source is empty and no use to requery.
                    log.debug(f'No data in {scenario}!')
                    self.row_queue.put('EMPTY')
                    return
                elif self.subscribe:
                    if status == SourceStatus.STREAMCLOSED:
                        log.debug(f'Subscribed data source {scenario} is active and not flagged hibernating, '
                                  f'but is not streaming')
                        self.row_queue.put('CLOSED')
                    elif status == SourceStatus.COMPLETED:
                        log.debug(f'Subscribed data source {scenario} has closed and stopped generating data')
                        self.row_queue.put('STOP')
                    elif status == SourceStatus.HIBERNATING:
                        log.debug(f'Subscribed data source {scenario} is hibernating')
                        self.row_queue.put('HIBERNATE')
                    return
                else:
                    # mode is continuous requry data and shift its _index to extend the current data
                    log.debug('Requerying data for '+scenario)
                    # TODO: Assess if we need to keep the query in memory instead of requery.
                    # Requery could lead to inconsistency perhaps? Data changed since last query?
                    self.generator = client.read_data_as_row(
                        tags=tags,
                        scenario=scenario, execution=execution,
                        start=self.start, end=self.end,
                        # We should never subscribe on requery
                        subscribe=False,
                        # Make sure to offset requeried _index so we extend the data. Shift by one estimated timestep
                        offset=self.max_ix_read+lastdt-self.start)

                if self.end > 0:
                    complete = self.max_ix_read >= self.end

            self.row_queue.put('STOP')
        threading.Thread(target=_next, daemon=True).start()

    def __next__(self):
        if not self.status == SourceStatus.OK:
            return self.status

        try:
            if self.last_get is None:
                self.last_get = time()
            row_ = self.row_queue.get(block=True, timeout=1)
            self.last_get = time()
        except Empty:
            row_ = None
            if self.timeout is not None and time() - self.last_get > self.timeout:
                log.debug('timeout')
                raise TimeoutError('Timed out!')

        if row_ == 'STOP':
            log.debug('source consumed and no longer active')
            self.status = SourceStatus.COMPLETED
            return self.status
        elif row_ == 'EMPTY':
            self.status = SourceStatus.EMPTYDATASET
            log.debug('Source empty')
            return self.status
        elif row_ == 'HIBERNATE':
            self.status = SourceStatus.HIBERNATING
            log.debug('Source is hibernating')
            return self.status
        elif row_ == 'CLOSED':
            self.status = SourceStatus.STREAMCLOSED
            log.debug('Source unexpectedly closed')
            return self.status
        elif row_ is None:
            return SourceStatus.NODATA

        row = {}
        for tag, s in self.spec['tags'].items():
            try:
                row[tag] = row_[s['tag']] * s['scale'] + s['offset']
            except Exception:
                log.debug(str(row_))
                raise
        # substract meta.offset since read_as_row returns the data in input base time
        row['_index'] = row_['_index'] + self.offset
        row['_index_relative'] = row_['_index'] - self.t0
        row['_datetime_utc'] = datetime.utcfromtimestamp(row['_index'])
        self.t = row['_index']
        self.row_buffer.add(row)
        return SourceStatus.OK

    # Get at simulation time
    def get_at_time(self, t):
        while self.t <= t or not self.row_buffer.full:
            status = self.__next__()
            if status != SourceStatus.OK:
                break

        if t > self.row_buffer.tmax():
            if self.status == SourceStatus.EMPTYDATASET:
                raise DataSourceEmptyDataset(f"Data Source for {self.name} is empty.")
            elif self.status == SourceStatus.HIBERNATING:
                raise DataSourceHibernating(f"Data Source for {self.name} is hibernating.")
            elif self.status == SourceStatus.COMPLETED:
                raise DataSourceCompleted(f"Data Source for {self.name} has been consumed and is no longer active.")
            elif self.status == SourceStatus.STREAMCLOSED:
                raise DataSourceStreamClosed(f"Data Source for {self.name} closed it's stream unexpectedly.")
            elif status == SourceStatus.NODATA:
                return None
            else:
                raise NotImplementedError(f"unhandled exception for {self.status}")
        else:
            return self.row_buffer.get(t)


class FIFO:
    def __init__(self, buffer_len=2):
        self.buffer = {}
        self.buffer_len = buffer_len

    def _init_buffer(self, row):
        self.buffer = {tag: np.zeros(self.buffer_len) for tag in row.keys()}

    def __len__(self):
        return len(self.buffer)

    def add(self, row):
        if len(self.buffer) == 0:
            self._init_buffer(row)

        for tag, val in row.items():
            self.buffer[tag] = np.roll(self.buffer[tag], -1)
            self.buffer[tag][-1] = val


class RowBuffer:
    def __init__(self, tags=None, offset=None, method=Interp.zero):
        self.tags = tags
        self.tags.append('_index')
        self.offset = offset
        self.method = method
        self.full = False
        self._rows = 0
        if method == Interp.zero:
            self.buffer_len = 2
        elif method == Interp.linear:
            self.buffer_len = 2
        else:
            raise NotImplementedError
        self._buffer = FIFO(self.buffer_len)

    def add(self, row):
        row_tags = {tag: row.get(tag) for tag in self.tags}
        self._buffer.add(row_tags)

        if not self.full:
            self._rows += 1
            if self._rows >= self.buffer_len:
                self.full = True

    def tmax(self):
        if "_index" in self._buffer.buffer:
            return max(self._buffer.buffer['_index'])
        else:
            return 0

    def get(self, t):
        tags = self.tags
        irow = {}
        if self.method == Interp.zero:
            for tag in tags[:-1]:
                ifun = interp1d(self._buffer.buffer['_index'], self._buffer.buffer[tag], kind='zero')
                irow.update({tag: ifun(t).tolist()})
        elif self.method == Interp.linear:
            for tag in tags[:-1]:
                ifun = interp1d(self._buffer.buffer['_index'], self._buffer.buffer[tag], kind='linear')
                irow.update({tag: ifun(t).tolist()})
        else:
            raise NotImplementedError('Unknown interpolation type')
        return irow


class InputManager:
    def __init__(self, sources, t, t0):
        self.sources = sources
        self.t = t
        self.t0 = t0

    def get_at_time(self, t):
        row = {}
        for source in self.sources:
            try:
                part_row = source.get_at_time(t)
                if part_row is None:
                    return None

            except TimeoutError as e:
                if source.subscribe:
                    return None
                else:
                    raise e
            row.update(part_row)
        # return simulation time
        row.update({'_index': t})
        return row

==> numerous_api_client-0.11.1b1.tar.gz: numerous_api_client-0.11.1b1/src/numerous_api_client/client/data_source.py <==
import enum
import threading
from datetime import datetime
from queue import Empty, Queue
from time import time
from typing import Any

import numpy as np
from scipy.interpolate import interp1d

from numerous_api_client.client.common import Interp, log


class DataSourceEmptyDataset(Exception):
    pass


class DataSourceHibernating(Exception):
    pass


class DataSourceCompleted(Exception):
    pass


class DataSourceStreamClosed(Exception):
    pass


class SourceStatus(enum.Enum):
    EMPTYDATASET = 0
    HIBERNATING = 1
    COMPLETED = 2
    STREAMCLOSED = 3
    OK = 4
    NODATA = 5


class StaticSource:
    def __init__(self, row_data, t0=0, t_end=None, dt=3600, generate_index=True):
        self.row = row_data
        self.source_type = 'static'
        self.t = np.inf
        self.t0 = t0
        self.subscribe = False
        self.generate_index = generate_index

    def get_at_time(self, t):
        self.t = t
        self.row['_index'] = t
        self.row['_index_relative'] = t - self.t0
        return self.row


class DynamicSource:
    def __init__(self, client: Any, scenario, execution, spec, t0=0, te=0, timeout=None, source_type=None,
                 interpolation_method=Interp.zero):
        self.timeout = timeout
        self.t0 = t0
        if te is None:
            te = 0
        self.te = te
        self.meta = client.get_timeseries_meta_data(scenario=scenario, execution=execution)
        self.closed = client.get_dataset_closed(scenario=scenario, execution=execution)
        self.name = scenario
        self.spec = spec
        self.spec['offset'] = self.spec['offset'] if 'offset' in self.spec else 0

        # The offset set in spec is the full offset from 0 - not relative from meta.offset
        offset = self.spec['offset']
        self.offset = offset
        self.execution = execution
        self.client = client

        # start time in original absolute time/input base time
        start = t0 - offset
        end = max(te - offset, 0)

        # If the data set is open, we will try to listen for data as its being generated if t end not reached.
        # This is stored in the state, to enable consistent behaviour across hibernation resumes
        subscribe = client.state.get(f"_{scenario}_{execution}_source_subscribe", None)
        if subscribe is None:
            subscribe = not self.closed
            client.state.set(f"_{scenario}_{execution}_source_subscribe", subscribe)
        self.subscribe = subscribe

        stats = client.get_timeseries_stats(None, scenario, execution)
        if stats.min + offset > t0:
            raise Exception(f'dynamic input scenario {scenario} with tags {[t.name for t in self.meta.tags]} '
                            f'is offset to the future {stats.min + offset}>{t0}')

        self.start = start
        self.end = end

        log.info('Scenario: %s', scenario)
        log.info('Scenario data is closed: %s', self.closed)
        log.info('Offset in spec: %s', datetime.fromtimestamp(spec['offset']))
        log.info('Offset in meta: %s', datetime.fromtimestamp(self.meta.offset))
        log.info('Data stats: %s', client.get_timeseries_stats(project=None, scenario=scenario, execution=execution))
        log.info('Combined offset: %s', datetime.fromtimestamp(offset))
        log.info('t0: %s', datetime.fromtimestamp(t0))
        log.info('te: %s', datetime.fromtimestamp(te))
        log.info('start: %s (%s)', datetime.fromtimestamp(start), start)
        log.info('end: %s (%s)', datetime.fromtimestamp(end), end)

        tags = set([s['tag'] for s in spec['tags'].values()])
        self.tags = tags

        self.generator = client.read_data_as_row(scenario=scenario, tags=tags, execution=execution, start=start,
                                                 end=end, subscribe=self.subscribe)
        self.source_type = source_type
        self.row_queue: Queue = Queue(100)
        self.t = -1
        self._rows_read = 0  # Keep a counter of rows read - if nothing is read we don't want to requery.
        self.row_buffer = RowBuffer(tags=[s for s in spec['tags'].keys()], offset=offset, method=interpolation_method)
        # Keep the last/presumably max ix read on record - if in continuous mode this needs to be  added to the requery
        # as an offset.
        self.max_ix_read = 0
        self.last_get = None
        self.status = SourceStatus.OK

        def _next():
            complete = False
            lastdt = 0
            lastrow_index = None
            status = SourceStatus.STREAMCLOSED

            while not complete:
                for row_ in self.generator:
                    if row_ == "HIBERNATE":
                        status = SourceStatus.HIBERNATING
                        break
                    elif row_ == "STOP":
                        status = SourceStatus.COMPLETED
                        break

                    self._rows_read += 1
                    self.row_queue.put(row_)
                    if lastrow_index is None:
                        lastrow_index = row_['_index']
                    lastdt = row_['_index']-lastrow_index
                    lastrow_index = row_['_index']
                    self.max_ix_read = row_['_index']

                log.debug("%s: All data read: %s of %d to %d, read n rows: %d", scenario, self.t-offset, start, end,
                          self._rows_read)

                if self._rows_read <= 0:
                    # If nothing was returned the data source is empty and no use to requery.
                    log.debug(f'No data in {scenario}!')
                    self.row_queue.put('EMPTY')
                    return
                elif self.subscribe:
                    if status == SourceStatus.STREAMCLOSED:
                        log.debug(f'Subscribed data source {scenario} is active and not flagged hibernating, '
                                  f'but is not streaming')
                        self.row_queue.put('CLOSED')
                    elif status == SourceStatus.COMPLETED:
                        log.debug(f'Subscribed data source {scenario} has closed and stopped generating data')
                        self.row_queue.put('STOP')
                    elif status == SourceStatus.HIBERNATING:
                        log.debug(f'Subscribed data source {scenario} is hibernating')
                        self.row_queue.put('HIBERNATE')
                    return
                else:
                    # mode is continuous requry data and shift its _index to extend the current data
                    log.debug('Requerying data for '+scenario)
                    # TODO: Assess if we need to keep the query in memory instead of requery.
                    # Requery could lead to inconsistency perhaps? Data changed since last query?
                    self.generator = client.read_data_as_row(
                        tags=tags,
                        scenario=scenario, execution=execution,
                        start=self.start, end=self.end,
                        # We should never subscribe on requery
                        subscribe=False,
                        # Make sure to offset requeried _index so we extend the data. Shift by one estimated timestep
                        offset=self.max_ix_read+lastdt-self.start)

                if self.end > 0:
                    complete = self.max_ix_read >= self.end

            self.row_queue.put('STOP')
        threading.Thread(target=_next, daemon=True).start()

    def __next__(self):
        if not self.status == SourceStatus.OK:
            return self.status

        try:
            if self.last_get is None:
                self.last_get = time()
            row_ = self.row_queue.get(block=True, timeout=1)
            self.last_get = time()
        except Empty:
            row_ = None
            if self.timeout is not None and time() - self.last_get > self.timeout:
                log.debug('timeout')
                raise TimeoutError('Timed out!')

        if row_ == 'STOP':
            log.debug('source consumed and no longer active')
            self.status = SourceStatus.COMPLETED
            return self.status
        elif row_ == 'EMPTY':
            self.status = SourceStatus.EMPTYDATASET
            log.debug('Source empty')
            return self.status
        elif row_ == 'HIBERNATE':
            self.status = SourceStatus.HIBERNATING
            log.debug('Source is hibernating')
            return self.status
        elif row_ == 'CLOSED':
            self.status = SourceStatus.STREAMCLOSED
            log.debug('Source unexpectedly closed')
            return self.status
        elif row_ is None:
            return SourceStatus.NODATA

        row = {}
        for tag, s in self.spec['tags'].items():
            try:
                row[tag] = row_[s['tag']] * s['scale'] + s['offset']
            except Exception:
                log.debug(str(row_))
                raise
        # substract meta.offset since read_as_row returns the data in input base time
        row['_index'] = row_['_index'] + self.offset
        row['_index_relative'] = row_['_index'] - self.t0
        row['_datetime_utc'] = datetime.utcfromtimestamp(row['_index'])
        self.t = row['_index']
        self.row_buffer.add(row)
        return SourceStatus.OK

    # Get at simulation time
    def get_at_time(self, t):
        while self.t <= t or not self.row_buffer.full:
            status = self.__next__()
            if status != SourceStatus.OK:
                break

        if t > self.row_buffer.tmax():
            if self.status == SourceStatus.EMPTYDATASET:
                raise DataSourceEmptyDataset(f"Data Source for {self.name} is empty.")
            elif self.status == SourceStatus.HIBERNATING:
                raise DataSourceHibernating(f"Data Source for {self.name} is hibernating.")
            elif self.status == SourceStatus.COMPLETED:
                raise DataSourceCompleted(f"Data Source for {self.name} has been consumed and is no longer active.")
            elif self.status == SourceStatus.STREAMCLOSED:
                raise DataSourceStreamClosed(f"Data Source for {self.name} closed it's stream unexpectedly.")
            elif self.status == SourceStatus.NODATA:
                return None
            else:
                raise NotImplementedError(f"unhandled exception for {self.status}")
        else:
            return self.row_buffer.get(t)


class FIFO:
    def __init__(self, buffer_len=2):
        self.buffer = {}
        self.buffer_len = buffer_len

    def _init_buffer(self, row):
        self.buffer = {tag: np.zeros(self.buffer_len) for tag in row.keys()}

    def __len__(self):
        return len(self.buffer)

    def add(self, row):
        if len(self.buffer) == 0:
            self._init_buffer(row)

        for tag, val in row.items():
            self.buffer[tag] = np.roll(self.buffer[tag], -1)
            self.buffer[tag][-1] = val


class RowBuffer:
    def __init__(self, tags=None, offset=None, method=Interp.zero):
        self.tags = tags
        self.tags.append('_index')
        self.offset = offset
        self.method = method
        self.full = False
        self._rows = 0
        if method == Interp.zero:
            self.buffer_len = 2
        elif method == Interp.linear:
            self.buffer_len = 2
        else:
            raise NotImplementedError
        self._buffer = FIFO(self.buffer_len)

    def add(self, row):
        row_tags = {tag: row.get(tag) for tag in self.tags}
        self._buffer.add(row_tags)

        if not self.full:
            self._rows += 1
            if self._rows >= self.buffer_len:
                self.full = True

    def tmax(self):
        if "_index" in self._buffer.buffer:
            return max(self._buffer.buffer['_index'])
        else:
            return 0

    def get(self, t):
        tags = self.tags
        irow = {}
        if self.method == Interp.zero:
            for tag in tags[:-1]:
                ifun = interp1d(self._buffer.buffer['_index'], self._buffer.buffer[tag], kind='zero')
                irow.update({tag: ifun(t).tolist()})
        elif self.method == Interp.linear:
            for tag in tags[:-1]:
                ifun = interp1d(self._buffer.buffer['_index'], self._buffer.buffer[tag], kind='linear')
                irow.update({tag: ifun(t).tolist()})
        else:
            raise NotImplementedError('Unknown interpolation type')
        return irow


class InputManager:
    def __init__(self, sources, t, t0):
        self.sources = sources
        self.t = t
        self.t0 = t0

    def get_at_time(self, t):
        row = {}
        for source in self.sources:
            try:
                part_row = source.get_at_time(t)
                if part_row is None:
                    return None

            except TimeoutError as e:
                if source.subscribe:
                    return None
                else:
                    raise e
            row.update(part_row)
        # return simulation time
        row.update({'_index': t})
        return row
